/*
 *  fp.c
 *  This program is prepared for the FINAL project for cs652
 *
 *  usage: fp [ -r lg_grid_resolution ] [ -p peaks_quantity ]
 *    ...for additional command line arguments,
 *    see the README file appended at or near the end of this
 *    source code listing.
 *
 *  written June 1996 by 	MIKE DINOLFO
 *
 *  modified August 1996:  add motif display/control options
 *
 *  Copyright 1996 by Mike Dinolfo
 *
 *  see the README.fp notes appended as comments at
 *  the end of this file listing
 */

/*  a typical compile statement for this program under linux:

    this statement revised July 7, 1998:

	cc -Wall fp.c -lMesatk -lMesaGLU -lMesaGL -lMesaaux \
                        -L/usr/X11/lib -I/usr/X11/include \
			-lm -lXt -lXext -o fp

    (Note- a typical Makefile for the site system is appended
       near the end of this source file.  However, as of July
        7, 1998, this Makefile is probably broken.  Use the
        command line compile statement given above.)

	NOTE: to compile with Motif capabilities, the command
		becomes:

        NOTE: as of July 7, 1998, this compile statement may
                well be broken.  Use at your own risk.

	cc -Wall fp.c -lMesatk -lMesaGLU -lMesaGL -lMesaaux \
			-lm -lXt -lXext -o fp \
			-DMOTIF -L/usr/X11R6/lib \
			-I/usr/X11R6/include -lXm -lXt -lX11 \
			-lXext
*/
#include <GL/gl.h>
#include <GL/glx.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
/*  This program is based on the method outlined in
  [Pickover 95] C. A. Pickover, "Generating Extraterrestrial
  Terrain", IEEE Computer Graphics and Applications, Vol 15,
  No 5, March 1995, pp. 18-21
 */
#include "glaux.h"

#ifdef MOTIF
#include <Xm/Xm.h>
#include <Xm/XmAll.h>
#include <Xm/Label.h>
#endif

#define delta_time 1.0
#define delta_heading 1.0
#define delta_speed 10.0
#define delta_altitude 100.

#define MIN_ALTITUDE 250.
#define MAX_ALTITUDE 6000.

#define SCREEN_SIZE 200

#define WORLD_SIZE 100000.

#define LG_GRID_RES_DEFAULT 5

#define LG_GRID_RES_MAXIMUM 8 /* must be = lg(GRID_RES_MAXIMUM) */
#define GRID_RES_MAXIMUM 256  /* must be = 2 ** LG_GRID_RES_MAXIMUM */

/*  these are the difficult ones to set:  */

#define HEIGHT_SCALE 1.5  /* default value! */
#define SPREAD_QTY 200    /* default value! */
#define SPREAD_WIDTH .0007  /* default value! */

#define DEFAULT_PEAKS_PER_GRID 20
/*
#define DEFAULT_PEAKS_PER_GRID 20
*/
#define MAX_PPG 1000

#define MAX_ELEVATION 6000. /* not utilized!! */

#define MAX_AIRSPEED 2000.

#define SHOW_TERRAIN_LIST 1

#define FOVY 40.
#define NEAR_Z_DIST 10.
#define FAR_Z_DIST 25000.

struct terdata  {
    int elevation;
    int highpoint;
    GLfloat nrml_magnitude[2];
};

#ifdef MOTIF
static int dblBuf[] = {
	GLX_DOUBLEBUFFER, GLX_RGBA, GLX_DEPTH_SIZE, 16,
	GLX_RED_SIZE, 1, GLX_GREEN_SIZE, 1, GLX_BLUE_SIZE, 1,
	None };
static int *snglBuf = &dblBuf[1];
XtAppContext app;
GLboolean doubleBuffer = GL_TRUE;
GLXContext cx;
#endif

struct terdata world[GRID_RES_MAXIMUM][GRID_RES_MAXIMUM];

/* global variables, related to aircraft position, speed, etc: */
GLfloat	airspeed,	/* just horizontal speed, units: meters/sec  */
	gpos_x,		/*   x position within the world (meters)	*/
	gpos_y,		/*   y     "       "   "   "      "             */
	altitude,	/* altitude above "sea level"; (meters)		*/
	heading,	/* compass orientation; 0.=north, 90.=east, etc */
	aspect_ratio;
int	grid_i,		/* current grid i position 			*/
	grid_j;		/*    "      "  j     "				*/

double height_scale, spread_width;
long spread_qty;

int z_buf_flag;

double navg_ht,nrms_ht;

GLfloat nhscale;

int peaks_per_grid;

int lg_grid_res;
int grid_res;

int i_start, j_start; /* limits for rendering */
int i_range, j_range;

double uvn(double xn)    {
    if (  ( xn > 1. )  ||  ( xn < 0. )  )
	fprintf(stderr,
	"check input to unv.\n");
    return (2./3.14159)*asin(sqrt(xn));
    }

double logit(double x)
    {
    return log(x/(1-x));
    }

void logistic_error(void)
{
	fprintf(stderr,"logistic error!\n");
	return;
}

int logistic(double lambda, int depth, double inval, double *nextval){
    if ( lambda <= 0. ) logistic_error();
    if ( lambda > 1. ) logistic_error();
    if ( (inval <= 0.) || (inval >= 1.) )
        *nextval = (rand()+1.)/(RAND_MAX + 2.);
    if ( depth > 0 )
        logistic (lambda, depth - 1, 4.*lambda*inval*(1.-inval),
			nextval);
    else
        *nextval = 4.*lambda*inval*(1.-inval);
    return 0;
}

void big_right_turn(void)    {
    /* calculate new heading:  */
    heading += 15;
    if ( heading >= 360. ) heading -= 360.;
    return;
    }

void big_left_turn(void)    {
    /* calculate new heading:  */
    heading -= 15;
    if ( heading < 0. ) heading += 360.;
    return;
    }

void hover(void)    {
    airspeed = 0.;
    return;
    }

void turn_left(void)   {
    heading -= delta_heading;
    if (heading < 0.) heading += 360.;
    return;
    }

void turn_right(void)   {
    heading += delta_heading;
    if(heading >= 360.)heading -= 360.;
    return;
    }

void increase_speed(void)    {
    airspeed += delta_speed;
    if (airspeed > MAX_AIRSPEED) 
		airspeed = MAX_AIRSPEED;
    return;
    }

void decrease_speed(void)    {
    airspeed -= delta_speed;
    if (airspeed < 0.) airspeed = 0.;
    return;
    }

void increase_altitude(void)    {
    altitude += delta_altitude;
/*
    if (altitude > MAX_ALTITUDE) altitude = MAX_ALTITUDE;
*/
    return;
    }

void decrease_altitude(void)    {
    altitude -= delta_altitude;
    if (altitude < MIN_ALTITUDE) altitude = MIN_ALTITUDE;
    return;
    }

void display(void);
void myReshape(GLsizei,GLsizei);

void update_aircraft_position(void)	{
    GLfloat angle;

    /* convert compass heading into angle:	*/
    angle = ( 90. - heading ) * 3.14159 / 180.;
    gpos_x += airspeed * delta_time * cos (angle);
    if (gpos_x >= WORLD_SIZE) gpos_x -= WORLD_SIZE;
    if (gpos_x < 0 ) gpos_x += WORLD_SIZE;
    gpos_y += airspeed * delta_time * sin (angle);
    if (gpos_y >= WORLD_SIZE) gpos_y -= WORLD_SIZE;
    if (gpos_y < 0 ) gpos_y += WORLD_SIZE;

    grid_i=(int)(grid_res*gpos_x/WORLD_SIZE);
/* presumably superfluous:
    if ( grid_i >= grid_res ) grid_i -= grid_res;
    if ( grid_i < 0 ) grid_i += grid_res;
*/
    grid_j=(int)(grid_res*gpos_y/WORLD_SIZE);
/* presumably superfluous:
    if ( grid_j >= grid_res ) grid_j -= grid_res;
    if ( grid_j < 0 ) grid_j += grid_res;
*/

    display();
    }

void myinit(void)
{
	GLfloat mat_amb[] = { 0.1, 1.0, 0.2, 1.0 };
	GLfloat mat_diff[] = { 0.1, 0.5, 0.2, 1.0 };
	glMaterialfv(GL_FRONT,GL_AMBIENT,mat_amb);  
	glMaterialfv(GL_FRONT,GL_DIFFUSE,mat_diff);
	if(z_buf_flag == 0)
		glEnable(GL_DEPTH_TEST);
	else
		glDisable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_FLAT);
	glColorMaterial(GL_FRONT,GL_DIFFUSE);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	glHint(GL_POLYGON_SMOOTH_HINT,GL_FASTEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_FASTEST);

	/* setup startup (initial) flight parameters:	*/

	airspeed = 1500.;	/* in meters/sec	*/
	gpos_x = WORLD_SIZE/2.;	/* origin coordinates	*/
	gpos_y = WORLD_SIZE/2.;
	altitude = (MAX_ALTITUDE + MIN_ALTITUDE) / 2.;
	heading = 80.;		/* east-northeast bound	*/
	grid_i = (int)(grid_res*(gpos_x/WORLD_SIZE));
	grid_j = (int)(grid_res*(gpos_y/WORLD_SIZE));

	nhscale = height_scale * 3.*32./
		(sqrt(peaks_per_grid)
			*grid_res);

	/* setup fog parameters: */
	{
		GLfloat fogColor[4] = { .25, .5, .75, 1. };
		GLint fogMode;
		glEnable(GL_FOG);
		fogMode = GL_LINEAR;
		glFogi (GL_FOG_MODE, fogMode);
		glFogfv (GL_FOG_COLOR, fogColor);
		glFogf (GL_FOG_START, NEAR_Z_DIST + 
			(FAR_Z_DIST - NEAR_Z_DIST) * .3 );
		glFogf (GL_FOG_END, NEAR_Z_DIST + 
			(FAR_Z_DIST - NEAR_Z_DIST) * 1.0 );
	glHint(GL_FOG_HINT,GL_FASTEST);
	glClearColor( .25,  .5, .75 , 1.0 ); 
			/*set sky-like blue background*/
/*  possible addition, but obviously needs "filling in": */
	/* set black sky at top to blue at bottom: */
/*
	glShadeModel(GL_SMOOTH);
	glBegin(GL_POLYGON);
		glColor3f(0.,0.,0.);
		glVertex3f(....);glVertex3f(....);
		glColor3f(0.,0.,1.);
		glVertex3f(....);glVertex3f(....);
	glEnd();
	glShadeModel(GL_FLAT);
*/
	}

	navg_ht=0.;
	nrms_ht=0.;

	return;
}

void do_display(void);
void do_nzb_display(void);

GLfloat normal_magnitude(GLfloat va[3], GLfloat vb[3], GLfloat vc[3])
{
	GLfloat vx1[3], vx2[3], vn[3], sumval;
	int i;
	for ( i =0; i < 3; i++ )
	{
		vx1[i]=va[i]-vb[i];
		vx2[i]=vb[i]-vc[i];
	}
	vn[0] = vx1[1]*vx2[2]-vx1[2]*vx2[1];
	vn[1] = vx1[0]*vx2[2]-vx1[2]*vx2[0];
	vn[2] = vx1[0]*vx2[1]-vx1[1]*vx2[0];
	sumval = 0.0;
	for ( i =0; i < 3; i++ ) sumval += vn[i] * vn[i];
	sumval = sqrt(sumval);
	return sumval;
}

void setup_world(int);

void generate_gridsquare_data(void)
{
	/* this routine uses logistics... */
	/* refer to Pickover article  */

	int i,j;
	long k;
	static double xx;

	/* we will want to do this only ONCE !! */
	xx = (rand()+1.)/(RAND_MAX+2.);

	/* initialize elevation values to zero: */
	for ( i = 0; i < grid_res ; i++ )
	{
	for ( j = 0; j < grid_res ; j++ )
		{
		world[i][j].elevation=0;
		}
	}

	/* we will generate a number of distinct peaks: */
	{
	double alpha, z, yn;
	int v, w;
	float real_cx,real_cy;

	int iter_limit = 10;
	for ( i = 0; i < peaks_per_grid; i++ )
		{
		logistic(1.,iter_limit,xx,&xx);
		alpha = uvn(xx);
		logistic(1.,iter_limit,xx,&xx);
		real_cx = uvn(xx) * grid_res;
		logistic(1.,iter_limit,xx,&xx);
		real_cy = uvn(xx) * grid_res;

fprintf(stdout,
	"Generating logistic data for peak no. %d of %d...\n",
		i+1,peaks_per_grid);

		for ( k = 0; k < 
			(long)
			  (alpha*alpha*
			  grid_res*grid_res*
			  spread_qty);
				k++)
			{
			logistic(1.,iter_limit,xx,&xx);
			z = (grid_res*spread_width)*logit(xx);
			logistic(1.,iter_limit,xx,&xx);
			yn = uvn(xx);
			if ( z > grid_res/2. ) continue;
			real_cx = 
			(z * alpha * sin (2*3.14159*yn) 
				+ real_cx );
			while (real_cx < 0.) 
				real_cx += grid_res;
			while (real_cx >= grid_res)
				real_cx -= grid_res;
			v = (int) real_cx;
			real_cy = 
			(z * alpha * cos (2*3.14159*yn) 
				+ real_cy );
			while (real_cy < 0.) 
				real_cy += grid_res;
			while (real_cy >= grid_res)
				real_cy -= grid_res;
			w = (int) real_cy;
			while ( v < 0 ) v += grid_res;
			while ( v >= grid_res ) v -= grid_res;
			while ( w < 0 ) w += grid_res;
			while ( w >= grid_res ) w -= grid_res;
			world[v][w].elevation++;
			}
		}
	}

	setup_world(grid_res);
	return;
}

void setup_world(int grid_res)
{
	int i,j,ii,jj;
	GLfloat v0[3],v1[3],v2[3],v3[3];
	float interval, offset;
	interval = WORLD_SIZE / grid_res;
	offset = 0.;
	for ( i =0; i < grid_res; i++ )
		{
		v0[0] = (i)*interval+offset;
		v1[0] = (i+1)*interval+offset;
		v2[0] = (i)*interval+offset;
		v3[0] = (i+1)*interval+offset;

		ii = i+1;
		if (ii == grid_res) ii = 0;

		for ( j = 0; j < grid_res; j++ )
		{
		v0[1]=(j)*interval+offset;
		v1[1]=(j)*interval+offset;
		v2[1]=(j+1)*interval+offset;
		v3[1]=(j+1)*interval+offset;

		jj = j+1;
		if (jj == grid_res) jj = 0;

		v0[2]=nhscale*world[i][j].elevation;
		v1[2]=nhscale*world[ii][j].elevation;
		v2[2]=nhscale*world[i][jj].elevation;
		v3[2]=nhscale*world[ii][jj].elevation;

			world[i][j].highpoint = world[i][j].elevation;
		if ( world[i][jj].elevation > world[i][j].highpoint )
			world[i][j].highpoint = world[i][jj].elevation;
		if ( world[ii][j].elevation > world[i][j].highpoint )
			world[i][j].highpoint = world[ii][j].elevation;
		if ( world[ii][jj].elevation > world[i][j].highpoint )
			world[i][j].highpoint = world[ii][jj].elevation;

		if (((i+j)/2)*2 == i+j)

		/*  EVEN: */
		{
			world[i][j].nrml_magnitude[0]= 
				normal_magnitude(v0, v1, v2);
			world[i][j].nrml_magnitude[1]= 
				normal_magnitude(v3, v2, v1);
		}
		else
		/*  ODD   */
		{
			world[i][j].nrml_magnitude[0]= 
				normal_magnitude(v2, v0, v3);
			world[i][j].nrml_magnitude[1]= 
				normal_magnitude(v1, v3, v0);
		}

		navg_ht += world[i][j].elevation;
		nrms_ht += world[i][j].elevation *
			world[i][j].elevation;
		}
		}

		navg_ht /= grid_res;
		navg_ht /= grid_res;
		nrms_ht /= grid_res;
		nrms_ht /= grid_res;
		nrms_ht = sqrt(nrms_ht);
	fprintf(stdout,"done with setup.  mean: %f, rms: %f\n",
		navg_ht,nrms_ht);
	return;
}

void generate_random_world(void)
{
	/* this routine uses rand(), not logistics */
	int i,j;
	long k;
	double xx;
	float real_cx,real_cy;

	/* initialize elevation values to zero: */
	for ( i = 0; i < grid_res ; i++ )
	{
	for ( j = 0; j < grid_res ; j++ )
		{
		world[i][j].elevation=0;
		}
	}

	/* we will generate a number of distinct peaks: */
	{
	double alpha, z, yn;
	int v, w;

	for ( i = 0; i < peaks_per_grid; i++ )
		{
		alpha = (rand()+1.)/(RAND_MAX + 2.);
		real_cx = (rand()+1.)/(RAND_MAX+2.)*
				grid_res;
		real_cy = (rand()+1.)/(RAND_MAX+2.)*
				grid_res;

fprintf(stdout,
	"Generating rand() data for peak no. %d of %d...\n",
		i+1,peaks_per_grid);

		for ( k = 0; k < 
			(long)
			  (alpha*alpha*grid_res*
			   grid_res*spread_qty);
				k++)
			{
			xx = sin((rand()+1.)*3.14159/
				(2.*(RAND_MAX+2.)));
			xx *= xx;
			z = (grid_res*spread_width*logit(xx));
			yn = (rand()+1.)/(RAND_MAX+2.);
			if ( z > grid_res/2. ) continue;
			real_cx = 
			(z * alpha * sin (2*3.14159*yn) 
				+ real_cx );
			while (real_cx < 0.) 
				real_cx += grid_res;
			while (real_cx >= grid_res)
				real_cx -= grid_res;
			v = (int) real_cx;
			real_cy = 
			(z * alpha * cos (2*3.14159*yn) 
				+ real_cy );
			while (real_cy < 0.) 
				real_cy += grid_res;
			while (real_cy >= grid_res)
				real_cy -= grid_res;
			w = (int) real_cy;
			
			while ( v < 0 ) v += grid_res;
			while ( v >= grid_res ) v -= grid_res;
			while ( w < 0 ) w += grid_res;
			while ( w >= grid_res ) w -= grid_res;
			world[v][w].elevation++;
			}
		}
	}

	setup_world(grid_res);
	return;
}

int mpaval(int *corner,int delta)
{
	int i,j;
	float newval;
	double xval;
	int possign;

	/* normalize indices to lie within
	 * our "world" limits: */
	for ( i=0; i<4; i++ )
	{
	  for ( j=0; j<2; j++ )
	  {
	  if(corner[i*2+j]>=grid_res)
	     corner[i*2+j]-=grid_res;
	  if(corner[i*2+j]<0)
	     corner[i*2+j]+=grid_res;
	  }
	}

	newval = 0.; /* initialize */

	for ( i=0; i<4; i++ )
	{
	   newval += 
	     world[corner[i*2]][corner[i*2+1]].elevation;  
		/* sum of 4 "corners" */
	}

	newval /= 4.0; /* average of 4 "corners" */

	/* here we need to compute a std. deviation
	 * based value; etc. */

	/* we start with a uniformly distributed
	 * random variable over the 0,1 range:
	 */
	xval = ((double)(rand()+1.))/(RAND_MAX+2.);

	/* we convert to u-shaped distribution.
	 *  according to Pickover, this will give
	 *  us a distribution of the form
	 *
	 *  P = 1 / (PI*sqrt(x * (1-x)))
	 */
	xval = sin(3.14159 * xval / 2.);
	xval = xval * xval;

	if (xval >= 0.5)possign = 1;
		else
			possign = -1;

	/* we now convert to near-Gaussian distribution: */
	xval = possign * log( xval / (1 - xval));

	return newval + (int)(100 * xval * delta);
}

void generate_mpa_world(int delta)
{

	int i,j,k,i2,j2;
	int corner[4][2];

	if(delta < 2) return;

	/*  initialize only one starting value */
	if(delta == grid_res)
		world[0][0].elevation = 0;

	for( i=0; i+delta/2<grid_res; i+=delta )
	{
 	  for( j=0; j+delta/2<grid_res; j+=delta )
	  {
	    corner[0][0] = i-delta/2;
	    corner[1][0] = i+delta/2;
	    corner[2][0] = i-delta/2;
	    corner[3][0] = i+delta/2;
	    corner[0][1] = j-delta/2;
	    corner[1][1] = j-delta/2;
	    corner[2][1] = j+delta/2;
	    corner[3][1] = j+delta/2;
	    world[i+delta/2][j+delta/2].elevation 
	        = mpaval(&corner[0][0],delta);
	  }
	}

	for( i=0; i+delta/2<grid_res; i+=delta )
	  {
	  for( j=0; j+delta/2<grid_res; j+=delta )
	  {
	    for( k=0; k<2; k++)
	    {
		if(k==0)
		{
			i2=i;
			j2=j+delta/2;
		}
		else
		{
			i2=i+delta/2;
			j2=j;
		}
	    corner[0][0] = i2;
	    corner[1][0] = i2-delta/2;
	    corner[2][0] = i2+delta/2;
	    corner[3][0] = i2;
	    corner[0][1] = j2-delta/2;
	    corner[1][1] = j2;
	    corner[2][1] = j2;
	    corner[3][1] = j2+delta/2;
	    world[i2][j2].elevation 
	       = mpaval(&corner[0][0],
			(int)(0.6+sqrt((float)delta)/2));
	    }
	  }
	}

	generate_mpa_world(delta/2);
	
	if ( delta <= 2 ) setup_world(grid_res);
	
	return;
}

void calculate_normal(GLfloat va[3], GLfloat vb[3], GLfloat vc[3], GLfloat *vn)
{
	GLfloat vx1[3], vx2[3], sumval;
	int i;
	for ( i =0; i < 3; i++ )
	{
		vx1[i]=va[i]-vb[i];
		vx2[i]=vb[i]-vc[i];
	}
	vn[0] = vx1[1]*vx2[2]-vx1[2]*vx2[1];
	vn[1] = vx1[0]*vx2[2]-vx1[2]*vx2[0];
	vn[2] = vx1[0]*vx2[1]-vx1[1]*vx2[0];
	sumval = 0.0;
	for ( i =0; i < 3; i++ ) sumval += vn[i] * vn[i];
	sumval = sqrt(sumval);
	for ( i =0; i < 3; i++ ) vn[i] /= sumval;
	return;
}

void consolidate_normal(GLfloat va[3], GLfloat vb[3], GLfloat vc[3], 
		GLfloat *vn,  GLfloat mag_value)
{
	GLfloat vx1[3], vx2[3];
	int i;
	for ( i =0; i < 3; i++ )
	{
		vx1[i]=va[i]-vb[i];
		vx2[i]=vb[i]-vc[i];
	}
	vn[0] = vx1[1]*vx2[2]-vx1[2]*vx2[1];
	vn[1] = vx1[0]*vx2[2]-vx1[2]*vx2[0];
	vn[2] = vx1[0]*vx2[1]-vx1[1]*vx2[0];
	for ( i =0; i < 3; i++ ) vn[i] /= mag_value;
	return;
}

void do_display(void);
void do_nzb_display(void);

void show_terrain(void)
{
	float angle_side,angle_start,angle_net;
	int i_min,i_max,j_min,j_max,i_temp,j_temp,n;

	float max_aa;

	/*  bump up altitude if necessary to clear
 	 *  hills within the viewer's gridsquare:
	 */
	max_aa = nhscale *
		world[grid_i][grid_j].highpoint;
	
	if(altitude <= max_aa )
	{
	fprintf(stdout,
	"       Increasing altitude to clear terrain!!\n");
  	 altitude = max_aa + 1.;
	}

	if((altitude > max_aa * 4. + 3000.) &&
		(peaks_per_grid >= 5))
	{
	fprintf(stdout,
	"       Decreasing altitude....\n");
		altitude *= 0.9;
	}

	/*  calculate the range of the "world" corresponding
	 *  to polygons which are within the field of view:
	 */
/*
	i_min = grid_i-1;
	i_max = grid_i+1;
	j_min = grid_j-1;
	j_max = grid_j+1;
*/
	i_min = grid_i-1;
	i_max = grid_i+1;
	j_min = grid_j-1;
	j_max = grid_j+1;

	/*  calculate angle to each side of line of sight to
	 *  far left/right corners of viewing world:
	 */

	/*  note: FOVY & heading in degrees, other angles in radians:  */

	angle_side = atan(tan(3.14159*FOVY/(2.*180.))*aspect_ratio);
	angle_start = (90.-heading)*3.14159/180.;

	for ( n=0; n<2; n++)
	{
	if(n == 0)angle_net = angle_start + angle_side;
	else	angle_net = angle_start - angle_side;
/*
	i_temp = grid_i + 
		(int)(grid_res*FAR_Z_DIST*cos(angle_net)/
			(WORLD_SIZE*cos(angle_side)));
	j_temp = grid_j + 
		(int)(grid_res*FAR_Z_DIST*sin(angle_net)/
			(WORLD_SIZE*cos(angle_side)));
*/
	i_temp = (int) ( (grid_res/WORLD_SIZE) *
			(gpos_x +
			FAR_Z_DIST*cos(angle_net)/
			cos(angle_side)));
	j_temp = (int) ( (grid_res/WORLD_SIZE) *
			(gpos_y +
			FAR_Z_DIST*sin(angle_net)/
			cos(angle_side)));

	if(i_temp -1 <= i_min) i_min = i_temp-1;
	if(i_temp +1 >= i_max) i_max = i_temp+1;
	if(j_temp -1 <= j_min) j_min = j_temp-1;
	if(j_temp +1 >= j_max) j_max = j_temp+1;
	}

	i_range = 1 + i_max - i_min;
	j_range = 1 + j_max - j_min;
	i_start = i_min;
	j_start = j_min;
	if( z_buf_flag == 0)
		do_display();
	else
		do_nzb_display();

	return;
}

void draw_poly(GLfloat *a0, GLfloat *a1, GLfloat *a2, 
	GLfloat *vn, GLfloat mag_value)
{
	consolidate_normal(a0,a1,a2,vn,mag_value);
	glNormal3fv(vn);
	glBegin(GL_TRIANGLES);
	glVertex3fv(a0);
	glVertex3fv(a1);
	glVertex3fv(a2);
	glEnd();
	return;
}

void do_display(void)
{
	int i,j,ii, jj,i_dummy,j_dummy;

	GLfloat mag_value;
	GLfloat interval, i_offset, j_offset;
	GLfloat v0[3],v1[3],v2[3],v3[3],vn[3];

	interval = WORLD_SIZE / grid_res;
/*
	if((heading <= 135.) && (heading >= 45.))
*/
	for ( i_dummy = i_start; i_dummy < i_start+i_range ; 
			i_dummy++ )
	{
	i = i_dummy;
	while ( i < 0 ) i += grid_res;
	while ( i >= grid_res ) i -= grid_res;

	i_offset = 0.;
	if ( (gpos_x-i*interval) > WORLD_SIZE/2. )
			i_offset = WORLD_SIZE;
	if ( (i*interval-gpos_x) > WORLD_SIZE/2. )
			i_offset = -WORLD_SIZE;
	v0[0] = (i)*interval+i_offset;
	v1[0] = (i+1)*interval+i_offset;
	v2[0] = (i)*interval+i_offset;
	v3[0] = (i+1)*interval+i_offset;

	ii = i+1;
	if ( ii == grid_res ) ii = 0;

	for ( j_dummy = j_start; j_dummy < j_start + j_range;
			j_dummy++ )
		{
		j = j_dummy;
		while ( j < 0 ) j += grid_res;
		while ( j >= grid_res ) j -= grid_res;
		j_offset = 0.;
	if ( (gpos_y-j*interval) > WORLD_SIZE/2. )
			j_offset = WORLD_SIZE;
	if ( (j*interval-gpos_y) > WORLD_SIZE/2. )
			j_offset = -WORLD_SIZE;
		v0[1]=(j)*interval+j_offset;
		v1[1]=(j)*interval+j_offset;
		v2[1]=(j+1)*interval+j_offset;
		v3[1]=(j+1)*interval+j_offset;

		jj = j+1;
		if ( jj == grid_res ) jj = 0;

		v0[2]=nhscale*world[i][j].elevation;
		v1[2]=nhscale*world[ii][j].elevation;
		v2[2]=nhscale*world[i][jj].elevation;
		v3[2]=nhscale*world[ii][jj].elevation;
		if((((i+j)/2)*2) != i+j)
		/*  ODD, upper left then lower right  */
			{
			mag_value = world[i][j].nrml_magnitude[0];
			draw_poly(v2,v0,v3,vn,
			  world[i][j].nrml_magnitude[0]);
			draw_poly(v1,v3,v0,vn,
			  world[i][j].nrml_magnitude[1]);
			}
		else /*   EVEN, lower left then upper right  */
			{
			draw_poly(v0,v1,v2,vn,
			  world[i][j].nrml_magnitude[0]);
			draw_poly(v3,v2,v1,vn,
			  world[i][j].nrml_magnitude[1]);
			}
		}
	}
	return;
}

void do_nzb_display(void)
{
	int i,j,ii, jj,i_dummy,j_dummy;

	int row_first;
	int toggle;

	int *first_v, *second_v;
	int first_start,first_end,second_start,second_end;
	int first_inc,second_inc;

	GLfloat interval, i_offset, j_offset;
	GLfloat v0[3],v1[3],v2[3],v3[3],vn[3];

	interval = WORLD_SIZE / grid_res;

	if((heading <= 135.) && (heading > 45.))
		{
		row_first = -1;

		first_v = &i_dummy;
		first_start = i_start+i_range;
		first_end = i_start;
		first_inc = -1;
		second_v = &j_dummy;
		second_start = j_start;
		second_end = j_start+j_range;
		second_inc = 1;
		}
	if((heading <= 45.) || (heading > 315.))
		{
		row_first = 1;
		first_v = &j_dummy;
		first_start = j_start+j_range;
		first_end = j_start;
		first_inc = -1;
		second_v = &i_dummy;
		second_start = i_start;
		second_end = i_start+i_range;
		second_inc = 1;
		}
	if((heading <= 315.) && (heading > 225.))
		{
		row_first = -1;

		first_v = &i_dummy;
		first_start = i_start;
		first_end = i_start+i_range;
		first_inc = 1;
		second_v = &j_dummy;
		second_start = j_start;
		second_end = j_start+j_range;
		second_inc = 1;
		}
	if((heading <= 225.) && (heading > 135.))
		{
		row_first = 1;

		first_v = &j_dummy;
		first_start = j_start;
		first_end = j_start+j_range;
		first_inc = 1;
		second_v = &i_dummy;
		second_start = i_start+i_range;
		second_end = i_start;
		second_inc = -1;
		}
	for ( *first_v = first_start; 
		(*first_v)*first_inc <first_end*first_inc;
		(*first_v)+=first_inc)
	{
	for(toggle = 0; toggle < 2; toggle++)
	{

	for ( *second_v = second_start; 
		(*second_v)*second_inc <second_end*second_inc;
		(*second_v)+=second_inc)

		{
		i = i_dummy;
		while ( i < 0 ) i += grid_res;
		while ( i >= grid_res ) i -= grid_res;
		i_offset = 0.;
		if ( (gpos_x-i*interval) > WORLD_SIZE/2. )
				i_offset = WORLD_SIZE;
		if ( (i*interval-gpos_x) > WORLD_SIZE/2. )
				i_offset = -WORLD_SIZE;
		v0[0] = (i)*interval+i_offset;
		v1[0] = (i+1)*interval+i_offset;
		v2[0] = (i)*interval+i_offset;
		v3[0] = (i+1)*interval+i_offset;
		ii = i+1;
		if ( ii == grid_res ) ii = 0;

		j = j_dummy;
		while ( j < 0 ) j += grid_res;
		while ( j >= grid_res ) j -= grid_res;
		j_offset = 0.;
	if ( (gpos_y-j*interval) > WORLD_SIZE/2. )
			j_offset = WORLD_SIZE;
	if ( (j*interval-gpos_y) > WORLD_SIZE/2. )
			j_offset = -WORLD_SIZE;
		v0[1]=(j)*interval+j_offset;
		v1[1]=(j)*interval+j_offset;
		v2[1]=(j+1)*interval+j_offset;
		v3[1]=(j+1)*interval+j_offset;

		jj = j+1;
		if ( jj == grid_res ) jj = 0;

		v0[2]=nhscale*world[i][j].elevation;
		v1[2]=nhscale*world[ii][j].elevation;
		v2[2]=nhscale*world[i][jj].elevation;
		v3[2]=nhscale*world[ii][jj].elevation;
	    if(toggle == 0)
	    {
		if((((i+j)/2)*2) != i+j)
		/*  ODD, upper left + lower right  */
			{
			if(((row_first == 1) &&
			   (first_inc == -1)) ||
			   ((row_first == -1) &&
			   (first_inc ==  1)))
			draw_poly(v2,v0,v3,vn,
			  world[i][j].nrml_magnitude[0]);
			else
			draw_poly(v1,v3,v0,vn,
			  world[i][j].nrml_magnitude[1]);
			}
		else /*   EVEN, upper right + lower left */
			{
			if(((row_first == 1) &&
			   (first_inc == -1)) ||
			   ((row_first == -1) &&
			   (first_inc == -1 )))
			draw_poly(v3,v2,v1,vn,
			  world[i][j].nrml_magnitude[1]);
			else
			draw_poly(v0,v1,v2,vn,
			  world[i][j].nrml_magnitude[0]);
			}
	    }
	    else
	    {
		if((((i+j)/2)*2) != i+j)
		/*  ODD, upper left + lower right  */
			{
			if(((row_first == 1) &&
			   (first_inc == 1)) ||
			   ((row_first == -1) &&
			   (first_inc ==  -1)))
			draw_poly(v2,v0,v3,vn,
			  world[i][j].nrml_magnitude[0]);
			else
			draw_poly(v1,v3,v0,vn,
			  world[i][j].nrml_magnitude[1]);
			}
		else /*   EVEN, upper right + lower left */
			{
			if(((row_first == 1) &&
			   (first_inc == 1)) ||
			   ((row_first == -1) &&
			   (first_inc == 1 )))
			draw_poly(v3,v2,v1,vn,
			  world[i][j].nrml_magnitude[1]);
			else
			draw_poly(v0,v1,v2,vn,
			  world[i][j].nrml_magnitude[0]);
			}
	    }
		}
	}
	}
	return;
}

void display(void)
{
	GLfloat light_orientation[] = { 0., 1000., 1000., 0. };
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );
	glColor3f(.0,1.0,.0);
	glLoadIdentity();
	gluLookAt(gpos_x,gpos_y,altitude,	/* where we are */
		gpos_x + sin(3.14159*heading/180.),
		gpos_y + cos(3.14159*heading/180.),
		altitude,	/* where we are looking at */
		0., 0., 1.);	/* "up" direction. */
	glEnable(GL_LIGHTING);
	glLightfv(GL_LIGHT0, GL_POSITION, light_orientation);
	glEnable(GL_LIGHT0);
	show_terrain();
	glFlush();
	glXSwapBuffers(auxXDisplay(), auxXWindow());

	fprintf(stdout,
	"hdg: %d, altitude: %d, xpos: %d, ypos: %d, ",
		(int)(heading + 0.5),(int)(altitude + 0.5),
		(int)(gpos_x + 0.5), (int)(gpos_y + 0.5)	);
	fprintf(stdout,
	"speed: %d\n",(int)(airspeed + 0.5));
		
	return;
}

void myReshape(GLsizei w, GLsizei h)
{
	glViewport(0, 0, w, h);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	aspect_ratio = (GLfloat) w / (GLfloat) h;

	gluPerspective(FOVY,  aspect_ratio,
		NEAR_Z_DIST,FAR_Z_DIST);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	glClearColor( .25,  .5, .75 , 1.0 ); /*set sky-like blue background*/
}

int main(int argc, char** argv)
{
	int method_id;
	int random_time_seed = 0;

	#ifdef MOTIF
	XtAppContext context;
	XmStringCharSet char_set=XmSTRING_DEFAULT_CHARSET;
	Widget toplevel, label;
	Display *dpy;
	XVisualInfo *vi;
	Arg al[10];
	int ac;
	#endif

	/* set default value(s):  */
	lg_grid_res = LG_GRID_RES_DEFAULT;
	peaks_per_grid = DEFAULT_PEAKS_PER_GRID;
	method_id = 1;
	spread_width = SPREAD_WIDTH;
	spread_qty = (long)SPREAD_QTY;
	height_scale = HEIGHT_SCALE;

	z_buf_flag = 1;

	/* parse/interpret command line flags:  */
	{
	int i;
	for ( i = 1; i < argc; i++ )
		{
		if ( argv[i][0] == '-')
			{
			switch ( argv[i][1] ) {
			case 'r': /* set grid resolution power */
				lg_grid_res = atoi(argv[++i]);
				if ( lg_grid_res <= 1 )
					lg_grid_res =
					LG_GRID_RES_DEFAULT;
				if ( lg_grid_res > LG_GRID_RES_MAXIMUM )
					lg_grid_res =
					LG_GRID_RES_DEFAULT;
				break;
			case 'f': /* user-fixed random seed */
				/* debugging only!! */
				random_time_seed = 1;
				if (i<argc-1)
				{
				  srand(atoi(argv[++i]));
				  break;
				}
				random_time_seed = 2;
				break;
			case 'm': /* user-defined method: */
				method_id = atoi(argv[++i]);
				break;
			case 'p': /* qty. of "mountain peaks" */
				peaks_per_grid = 
				  atoi(argv[++i]);
				if((peaks_per_grid < 1) ||
				   (peaks_per_grid > 
					MAX_PPG))
				  peaks_per_grid =
				   DEFAULT_PEAKS_PER_GRID;
				break;
			case 'w': /* define spread_width */
				spread_width =
				  atof(argv[++i]);
				break;
			case 'q': /* define spread_qty */
				spread_qty =
				  (long)atoi(argv[++i]);
				break;
			case 'h': /* define height_scale */
				height_scale =
				  atof(argv[++i]);
				break;
			case 'z': 
				z_buf_flag = 0;
				break;
			default:
fprintf(stderr,
"usage: fp [-r lg_grid_resolution][-p peaks_quantity]\n");
fprintf(stderr,
"Additional parameters are described in the README file.\n");
				exit(0);
			}  }
		else
			{
			fprintf(stderr,"com line lacks '-' for flag!\n");
			exit(0);
			}
		}
	}

	{
	int i;
	grid_res = 1;
	for ( i = 0; i < lg_grid_res ; i++ ) grid_res *= 2;
	}

	if ( random_time_seed == 0 )
	{
	time_t timevalue;

	if ( time(&timevalue) == -1 )
		{
		fprintf(stdout,
		  "unable to access system clock for seed.\n");
		}
	else
		{
		long int i;
		struct tm *clockreading;
		clockreading = localtime(&timevalue);
	/* note- in the following we are reversing significant
	 * bit ordering; this is intentional:  */
		i =
		    (*clockreading).tm_yday +
		    (*clockreading).tm_hour *60 +
		    (*clockreading).tm_min *60*60 +
		    (*clockreading).tm_sec *24*60*60;
		srand( i );
		}
	}

/* DEBUG: */
/*
generate_gridsquare_data(lg_grid_res);
*/

	auxInitDisplayMode (AUX_DEPTH | AUX_DOUBLE | AUX_RGB);

	auxInitPosition (0, 0, SCREEN_SIZE, SCREEN_SIZE);
	auxInitWindow (argv[0]);

	#ifdef MOTIF
	toplevel=XtAppInitialize(&context,"",NULL,0,&argc,
		argv,NULL,NULL,0);
	ac=0;
	XtSetArg(al[ac],XmNlabelString,
	  XmStringCreateLtoR("Hello world",char_set)); ac++;
	dpy = XtDisplay(toplevel);
	
	vi = glXChooseVisual(dpy,DefaultScreen(dpy),dblBuf);
	if (vi == NULL) {
	    vi = glXChooseVisual(dpy,DefaultScreen(dpy),snglBuf);
		if ( vi == NULL)
		    XtAppError(app,
			"no RGB visual with depth buffer");
		doubleBuffer = GL_FALSE;
	}
	cx = glXCreateContext(dpy,vi, /* no display list sharing */
		None, /* favor direct */ GL_TRUE);
	if ( cx == NULL)
	    XtAppError(app,"could not create rendering context");

	label=XmCreateLabel(toplevel,"label",al,ac);
	#endif

	myinit();

	auxKeyFunc(AUX_LEFT, turn_left);
	auxKeyFunc(AUX_RIGHT, turn_right);
	auxKeyFunc(AUX_UP, increase_speed);
	auxKeyFunc(AUX_DOWN, decrease_speed);
	auxKeyFunc(AUX_u,  increase_altitude);
	auxKeyFunc(AUX_U,  increase_altitude);
	auxKeyFunc(AUX_d,  decrease_altitude);
	auxKeyFunc(AUX_D,  decrease_altitude);

	auxKeyFunc(AUX_l,  big_left_turn);
	auxKeyFunc(AUX_L,  big_left_turn);
	auxKeyFunc(AUX_r,  big_right_turn);
	auxKeyFunc(AUX_R,  big_right_turn);
	auxKeyFunc(AUX_s,  hover);
	auxKeyFunc(AUX_S,  hover);

	auxReshapeFunc(myReshape); 

	auxIdleFunc(update_aircraft_position);

	switch (method_id)  {
	  case 0:generate_random_world();
	         break;
	  case 1:generate_random_world();
		 break;
	  case 2:generate_mpa_world(grid_res);
		 break;
	  default:
		 fprintf(stdout,"bad method_id var!\n");
		 exit(0);
	}
/*
	if ( method_id != 1 )
		generate_gridsquare_data(lg_grid_res);
	else
		generate_random_world(lg_grid_res);
*/

	#ifdef MOTIF
	XtManageChild(label);
	XtRealizeWidget(toplevel);
	#endif

	auxMainLoop(display);

	#ifdef MOTIF
	XtAppMainLoop(context);
	#endif

	return 0;
}
/* the following is the Makefile for the site system
 * to compile, type 'make targets'
 *
 *  NOTE: no motif capability in this makefile!
 *
 *  NOTE: as of July 7, 1998, this makefile is probably
 *    NO LONGER VALID.  I suggest that you go to the top
 *    of this source code file; you will find a compile statement
 *    right after the #include statements which should work OK.
 *
##### MACROS #####
CC = gcc

CFLAGS = -O3 -I/usr/openwin/include -DSHM

INCDIR = /opt/Mesa-1.2.7/include

GL_LIBS = -L/opt/Mesa-1.2.7/lib -R/opt/Mesa-1.2.7/lib -lMesaaux -lMesatk -lMesaGLU -lMesaGL -lm -L/usr/openwin/lib -lX11 -lXext

LIB_DEP = /opt/Mesa-1.2.7/lib/$(GL_LIB) /opt/Mesa-1.2.7/lib/$(GLU_LIB) /opt/Mesa-1.2.7/lib/$(TK_LIB) /opt/Mesa-1.2.7/lib/$(AUX_LIB)

PROGS = fp

##### RULES #####

.SUFFIXES:
.SUFFIXES: .c

.c: $(LIB_DEP)
	$(CC) -I$(INCDIR) $(CFLAGS) $< $(GL_LIBS) -o $@

##### TARGETS #####

default:
	@echo "Specify a target configuration"

clean:
	-rm *.o *~

realclean:
	-rm $(PROGS)
	-rm *.o *~

targets: $(PROGS)
			end of site Makefile */
/*  the README.fp file:

These are brief, although complete, instructions for
running the "fp" program.  Also, be sure to read
the "BUGS" portion.

OPENING THE CORRECT WINDOWS IN THE XWINDOW ENVIRONMENT:

The program produces a considerable amount of
textual information as it is running- this text output
is sent to stdout.  As a result, it is necessary to have
the default shell window open and visible, preferrably in the
lower right hand corner of the video screen.  The program's
graphics output window is likely to be located in the upper
left hand corner; placement of the shell window in the lower
right will allow the user to view both the graphics (in the
upper left) and the text (in the lower right).

Once you understand this point, all that you have to do to 
start the program is to type "fp" to start it running.  The
program will begin producing numeric data to define the 
terrain; as soon as all of the numeric data is produced,
the program will commence producing graphic output.  

There are a limited number of OPTIONAL command line 
arguments; the complete format for the command line is

	fp [ -r lg_grid_resolution ] [ -p peaks_quantity ]
	   [ -m method_id ] [ -q spread_quantity ]
           [ -h height_scale ] [ -w spread_width ]
	   [ -z  ]

I _RECOMMEND_ that you initially run the program with 
no arguments, by typing the following line:

         fp

this seems to generate an acceptable display.  Certain
values for command line arguments result in a less
natural simulation of terrain.  However, a description of
the command line arguments is as follows:

'Lg_grid_resolution' (integer) is the log(base 2) of the grid 
size utilized for generating terrain.  Minimum value of 
'lg_grid_resolution' is 1, corresponding to a 2 x 2 grid;
maximum value of 'lg_grid_resolution' is 8, corresponding to a
256 x 256 grid; default value of 'lg_grid_resolution' is
5, corresponding to a 32 x 32 grid. 

'Peaks_quantity' is the (integer) quantity of "peaks" for 
terrain feature generation.  Minimum value is 2, maximum 
value is 1000, default value is 20. 

'Method_id' is an integer variable to define the method for
generation of random terrain.  If set equal to 0 (zero),
then a logistic method will be utilized (per Pickover's 
article).  If set equal to 1, an (apparently equivalent) 
random method will be utilized (via the rand() function) 
to generate essentially equivalent terrain.  The default 
value is 1.

'-z' is a flag to implement the z-buffer as a means of
hidden surface removal.  I recommend that you do NOT set
this flag- the default performance will result in faster
rendering via a software-implemented hidden surface
removal algorithm.

'height_scale' (float) is a scaling factor which is 
multiplied times the integer value at each grid peak to 
determine the terrain elevation. 

'spread_qty' (integer) defines the number of points which are 
generated for each peak.

'spread_width' (float) is a value which is multiplied times the
radius of each point for each peak to determine its position.

RECOMMENDATION:  use the default values for height_scale,
spread_qty, and spread_width unless necessary to provide realistic
terrain.

An attempt to set any of these user-defined variables to
values outside of the allowable limits will result in setting
them to their default values.  As you might expect, setting
either lg_grid_resolution or peaks_quantity  to larger values 
results in increases in terrain data generation time and/or 
rendering time.

USER INPUT:

Pressing the LEFT ARROW key or RIGHT ARROW key will turn the
simulated aircraft left or right.  Each keypress will result
in a one degree change in heading.  Larger turns may be
executed by repeatedly pressing the left/right arrow keys,
or by pressing the 'l' key (which gives a 15 degree turn 
to the left) or by pressing the 'r' key (which gives a 15 
degree turn to the right).

Pressing the 'u' key will increase the simulated aircraft's
altitude (corresponding to "up" movement).  Pressing the
'd' key will decrease the simulated aircraft's altitude
("down" movement).

Pressing the 's' key will instantly STOP the aircraft's
motion, reducing the speed to zero.  This is useful if it
is desired to change direction; the aircraft can be stopped,
then rotated to the desired heading, and then the speed
can slowly be increased from zero to whatever speed is
desired.

NOTE:  as the simulated aircraft approaches hills whose 
peak elevation is greater than that of the simulated aircraft,
the program will INCREASE the aircraft's altitude to avoid
"collision" with the hill.  As a result, it may be
necessary for the user to press the 'd' key to DECREASE the 
altitude after flying "over" such a hill.  This will bring the
simulated aircraft's altitude down sufficiently to permit
viewing of the terrain.  However, the program will also act to
DECREASE the aircraft's altitude after flying over a peak, so
that the altitude is somewhat "self-regulating".

Pressing the UP ARROW key will increase speed; pressing the
DOWN ARROW key will decrease speed.  It is possible to decrease
speed to zero, in which case the simulated aircraft will
"hover" above the ground.  Speed can also be instantaneously
decreased to zero by pressing the 's' key (for "stop").

BUGS

Sometimes, when peak data is generated using the logistic
method (i.e., when the program is run with the
-m command line argument), the program will result in
a fatal floating point exception error.  I believe that this
is due to an overflow or underflow condition in the 
iterative number generator.  As a result, I recommend that
all runs by performed without the '-m' command line argument.


*/
